
from collections import namedtuple
from estime2.helper import raise_if_not_subset
import re



RegisteredOption = namedtuple(
    "RegisteredOption", 
    "key defval doc validator"
)
age_down_to = 40
age_floor = 50
age_mask = []
age_max = 99
age_prop_size = 5
comp_neg_temp_out = 'TEM'
comp_neg_emi = 'EMI'
comp_neg_npr_out = 'NPR, 2018-07-01'
comp_neg_death = 'DTH'
comp_neg_interprov_out = 'IOM'
comp_neg_intraprov_out = None
comp_neg_etc = ['RAO']
comp_neg_put_etc_before = False
comp_pos_temp_in = None
comp_pos_ret_emi = 'RE'
comp_pos_npr_in = 'NPR, 2019-07-01'
comp_pos_immi = 'IMM'
comp_pos_interprov_in = 'IIM'
comp_pos_intraprov_in = None
comp_pos_etc = ['RAI']
comp_pos_put_etc_before = False
comp_end = [comp_pos_npr_in]
method_all = ['1dist', 'filler', 'prop']
method = '1dist'
method_use_age_floor = True
method_use_old_neigh = False
method_use_second_pass = True
method_use_seq = True
pop_age = 'Age'
pop_birth = 'BTH'
pop_at_least = 1
pop_end = 'Postcensal Population'
pop_sex = 'Sex'
pop_start = 'Initial Population'



def is_type(x, opname, supposed_type_x):
    if not isinstance(x, supposed_type_x):
        raise TypeError(
            f"{opname} must be {supposed_type_x}, not of {type(x)}."
        )
    return None

def is_bool(x, opname):
    return is_type(x, opname, bool)

def is_int(x, opname):
    return is_type(x, opname, int)

def is_str(x, opname, can_be_None = True):
    if x is None:
        if can_be_None:
            return None
        else:
            raise TypeError(f"{opname} cannot be None; it must be a str.")
    return is_type(x, opname, str)

def is_nonnegative(x, opname, force_positive = False):
    is_type(x, opname, int)
    x_test = None
    if force_positive:
        x_test = (x <= 0)
    else:
        x_test = (x < 0)
    if x_test:
        raise ValueError(
            f"{opname} must be a nonnegative integer, not {x}."
        )
    return None

def is_list_of_item_type(x, opname, item_type):
    is_type(x, opname, list)
    for i, item in enumerate(x):
        if not isinstance(item, item_type):
            raise ValueError(
                f"{opname} has an item at index {i} that is not of "
                f"type {item_type} but of {type(item)}."
            )
    return None

def is_list_containing(x, opname, check_subset: bool = True):
    is_list_of_item_type(x, opname, str)
    if check_subset:
        comps =\
            [
                comp_neg_temp_out, comp_neg_emi, comp_neg_npr_out, 
                comp_neg_death, 
                comp_neg_interprov_out, comp_neg_intraprov_out
            ] +\
                comp_neg_etc +\
            [
                comp_pos_temp_in, comp_pos_ret_emi, comp_pos_npr_in,
                comp_pos_immi, 
                comp_pos_interprov_in, comp_pos_intraprov_in
            ] +\
                comp_pos_etc
        if not set(x).issubset(comps):
            raise ValueError(
                f"{x} is not a subset of all the components."
            )

    return None

def is_method(x):
    is_str(x, 'method', False)
    raise_if_not_subset([x], method_all, x, str(method_all))

    return None

is_bool_afloor = lambda x: is_bool(x, 'method_use.age_floor')
is_bool_mu2p = lambda x: is_bool(x, 'method_use.second_pass')
is_bool_old_neigh = lambda x: is_bool(x, 'method_use.old_neigh')
is_bool_seq = lambda x: is_bool(x, 'method_use.seq')
is_containing_comp_end = lambda x: is_list_containing(x, 'comp.end', False)
is_containing_netc = lambda x: is_list_containing(x, 'comp_neg.etc', False)
is_containing_petc = lambda x: is_list_containing(x, 'comp_pos.etc', False)
is_etc_n_b4_inters = lambda x: is_bool(x, 'comp_neg.put_etc_before')
is_etc_p_b4_inters = lambda x: is_bool(x, 'comp_pos.put_etc_before')
is_int_age_max = lambda x: is_int(x, 'age.max')
is_int_age_down_to = lambda x: is_int(x, 'age.down_to')
is_int_age_floor = lambda x: is_int(x, 'age.floor')
is_list_str_age_mask = lambda x: is_list_of_item_type(x, 'age.mask', str)
is_pos_age_prop_size = lambda x: is_nonnegative(x, 'age.prop_size', True)
is_pos_pop_at_least = lambda x: is_nonnegative(x, 'pop.at_least')
is_str_comp_neg_temp_out = lambda x: is_str(x, 'comp_neg.temp_out')
is_str_comp_neg_emi = lambda x: is_str(x, 'comp_neg.emi')
is_str_comp_neg_npr_out = lambda x: is_str(x, 'comp_neg.npr_out')
is_str_comp_neg_death = lambda x: is_str(x, 'comp_neg.death')
is_str_comp_neg_interprov_out = lambda x: is_str(x, 'comp_neg.interprov_out')
is_str_comp_neg_intraprov_out = lambda x: is_str(x, 'comp_neg.intraprov_out')
is_str_comp_pos_temp_in = lambda x: is_str(x, 'comp_pos.temp_in')
is_str_comp_pos_ret_emi = lambda x: is_str(x, 'comp_pos.ret_emi')
is_str_comp_pos_npr_in = lambda x: is_str(x, 'comp_pos.npr_in')
is_str_comp_pos_immi = lambda x: is_str(x, 'comp_pos.immi')
is_str_comp_pos_interprov_in = lambda x: is_str(x, 'comp_pos.interprov_in')
is_str_comp_pos_intraprov_in = lambda x: is_str(x, 'comp_pos.intraprov_in')
is_str_pop_age = lambda x: is_str(x, 'pop.age', False)
is_str_pop_birth = lambda x: is_str(x, 'pop.birth', False)
is_str_pop_end = lambda x: is_str(x, 'pop.end', False)
is_str_pop_sex = lambda x: is_str(x, 'pop.sex', False)
is_str_pop_start = lambda x: is_str(x, 'pop.start', False)



_global_config = {
    'age': {
        'down_to': age_down_to,
        'floor': age_floor,
        'mask': age_mask,
        'max': age_max,
        'prop_size': age_prop_size
    },
    'comp': {
        'end': comp_end
    },
    'comp_neg': {
        'temp_out': comp_neg_temp_out,
        'emi': comp_neg_emi,
        'npr_out': comp_neg_npr_out,
        'death': comp_neg_death,
        'interprov_out': comp_neg_interprov_out,
        'intraprov_out': comp_neg_intraprov_out,
        'etc': comp_neg_etc,
        'put_etc_before': comp_neg_put_etc_before
    },
    'comp_pos': {
        'temp_in': comp_pos_temp_in,
        'ret_emi': comp_pos_ret_emi,
        'npr_in': comp_pos_npr_in,
        'immi': comp_pos_immi,
        'interprov_in': comp_pos_interprov_in,
        'intraprov_in': comp_pos_intraprov_in,
        'etc': comp_pos_etc,
        'put_etc_before': comp_pos_put_etc_before
    },
    'method': method,
    'method_use': {
        'age_floor': method_use_age_floor,
        'old_neigh': method_use_old_neigh,
        'second_pass': method_use_second_pass,
        'seq': method_use_seq
    },
    'pop': {
        'age': pop_age,
        'birth': pop_birth,
        'at_least': pop_at_least,
        'end': pop_end,
        'sex': pop_sex,
        'start': pop_start
    }
}
_registered_options = {
    'age.down_to': RegisteredOption(
        key = 'age.down_to',
        defval = age_down_to,
        doc =\
            "\n (int)\n    " +\
            "(40 by default) A limit to which neighbouring ages are " +\
            "considered by the estime2 package. Counter-adjustments " +\
            "are applied from the maximum neighbouring age and " +\
            "all the way down to min(age.floor, max(-1, " +\
            "age of modification - `age.down_to`)) if " +\
            "`method_use.age_floor` is True, or max(-1," +\
            "age of modification - `age.down_to`) if " +\
            "`method_use.age_floor` is False.",
        validator = is_int_age_down_to
    ),
    'age.floor': RegisteredOption(
        key = 'age.floor',
        defval = age_floor,
        doc =\
            "\n (int)\n    " +\
            "(50 by default) A bottom level of old neighbouring ages " +\
            "of the problematic age. If `method_use.age_floor` is " +\
            "True, then the minimum counter-adjusted neighbouring age " +\
            "will be either `age.floor` or max(-1, age of modification " +\
            "- `age.down_to`), whichever is smaller. If " +\
            "`method_use.age_floor` is False, then it will be " +\
            "max(-1, age of modification - `age.down_to`) and this " +\
            "option will be ignored.",
        validator = is_int_age_floor
    ),
    'age.mask': RegisteredOption(
        key = 'age.mask',
        defval = age_mask,
        doc =\
            "\n (list of str)\n    " +\
            '(`[]` by default) A list of ages to "mask" or "ignore" ' +\
            'in the end-of-period population table so that when these ' +\
            'ages happen to be one of neighbouring ages of a certain ' +\
            'problematic record, their end-of-period populations are ' +\
            'considered to be 0.',
        validator = is_list_str_age_mask
    ),
    'age.max': RegisteredOption(
        key = 'age.max',
        defval = age_max,
        doc =\
            "\n (int)\n    " +\
            '(99 by default) A maximum individual age. Any age bigger ' +\
            'than this option is classified as the "plus category." ' +\
            'That is, if `age.max` == 99, then any age greater than or' +\
            'equal to 100 is classified as `100+`.', 
        validator = is_int_age_max
    ),
    'age.prop_size': RegisteredOption(
        key = 'age.prop_size',
        defval = age_prop_size,
        doc =\
            "\n (int)\n    " +\
            '(5 by default) A size of period to use when applying the ' +\
            'proportional method or the sequential one-distribution ' +\
            '(see `method_use.seq`). It must be a positive integer.',
        validator = is_pos_age_prop_size
    ),
    'comp.end': RegisteredOption(
        key = 'comp.end',
        defval = comp_end,
        doc =\
            '\n (list of str)\n '+\
            'A list of components that denote the end-of-period ' +\
            'figures. This must be a subset of all the components.',
        validator = is_containing_comp_end
    ),
    'comp_neg.temp_out': RegisteredOption(
        key = 'comp_neg.temp_out',
        defval = comp_neg_temp_out,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Temporary emigrants OUT" ' +\
            "during the period in a population table. ",
        validator = is_str_comp_neg_temp_out
    ),
    'comp_neg.emi': RegisteredOption(
        key = 'comp_neg.emi',
        defval = comp_neg_emi,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Emigrants" ' +\
            "during the period in a population table.",
        validator = is_str_comp_neg_emi
    ),
    'comp_neg.npr_out': RegisteredOption(
        key = 'comp_neg.npr_out',
        defval = comp_neg_npr_out,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Non-permanent residents" (OUT) ' +\
            "at the end of the period in a population table.",
        validator = is_str_comp_neg_npr_out
    ),
    'comp_neg.death': RegisteredOption(
        key = 'comp_neg.death',
        defval = comp_neg_death,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Deaths" ' +\
            "during the period in a population table.",
        validator = is_str_comp_neg_death
    ),
    'comp_neg.interprov_out': RegisteredOption(
        key = 'comp_neg.interprov_out',
        defval = comp_neg_death,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Interprovincial migration OUT" ' +\
            "during the period in a population table.",
        validator = is_str_comp_neg_interprov_out
    ),
    'comp_neg.intraprov_out': RegisteredOption(
        key = 'comp_neg.intraprov_out',
        defval = comp_neg_death,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Intraprovincial migration OUT" ' +\
            "during the period in a population table.",
        validator = is_str_comp_neg_intraprov_out
    ),
    'comp_neg.etc': RegisteredOption(
        key = 'comp_neg.etc',
        defval = comp_neg_etc,
        doc =\
            "\n (list of str)\n " +\
            "A list of negative components that is not a part of " +\
            "conventional components. Components in this list is " +\
            "considered after all the conventional negative components " +\
            "have been dealt with. If no extra negative components " +\
            "exist in the table, let this be `[]`, an empty list.",
        validator = is_containing_netc
    ),
    'comp_neg.put_etc_before': RegisteredOption(
        key = 'comp_neg.put_etc_before',
        defval = comp_neg_put_etc_before,
        doc =\
            "\n (bool)\n " +\
            "(`False` by default) A bool that determines whether " +\
            "or not components in `comp_neg.etc` should precede " +\
            "`comp_neg.interprov` and `comp_neg.intraprov` when " +\
            "applying corrections. If `True`, the precedence of " +\
            "components in `comp_neg.etc` will be between " +\
            "conventional negative components and interprovincial " +\
            "migrations. If `False`, it is " +\
            "set to follow interprovincial or intraprovincial " +\
            "migration, depending on the regional level of the " +\
            "population table.",
        validator = is_etc_n_b4_inters
    ),
    'comp_pos.temp_in': RegisteredOption(
        key = 'comp_pos.temp_in',
        defval = comp_pos_temp_in,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Temporary emigrants IN" ' +\
            "during the period in a population table.",
        validator = is_str_comp_pos_temp_in
    ),
    'comp_pos.ret_emi': RegisteredOption(
        key = 'comp_pos.ret_emi',
        defval = comp_pos_ret_emi,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Returning emigrants" ' +\
            "during the period in a population table.",
        validator = is_str_comp_pos_ret_emi
    ),
    'comp_pos.npr_in': RegisteredOption(
        key = 'comp_pos.npr_in',
        defval = comp_pos_npr_in,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Non-permanent residents" (IN) ' +\
            "at the start of the period in a population table.",
        validator = is_str_comp_pos_ret_emi
    ),
    'comp_pos.immi': RegisteredOption(
        key = 'comp_pos.immi',
        defval = comp_pos_immi,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Immigrants" ' +\
            "during the period in a population table.",
        validator = is_str_comp_pos_immi
    ),
    'comp_pos.interprov_in': RegisteredOption(
        key = 'comp_pos.interprov_in',
        defval = comp_pos_interprov_in,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Interprovincial migration IN" ' +\
            "during the period in a population table.",
        validator = is_str_comp_pos_interprov_in
    ),
    'comp_pos.intraprov_in': RegisteredOption(
        key = 'comp_pos.intraprov_in',
        defval = comp_pos_interprov_in,
        doc =\
            "\n (str or None)\n " +\
            'A name that denotes "Intraprovincial migration IN" ' +\
            "during the period in a population table.",
        validator = is_str_comp_pos_intraprov_in
    ),
    'comp_pos.etc': RegisteredOption(
        key = 'comp_pos.etc',
        defval = comp_pos_etc,
        doc =\
            "\n (list)\n " +\
            "A list of positive components that is not a part of " +\
            "conventional components. Components in this list is " +\
            "considered after all the conventional positive components " +\
            "have been dealt with. If no extra positive components " +\
            "exist in the table, let this be `[]`, an empty list.",
        validator = is_containing_petc
    ),
    'comp_pos.put_etc_before': RegisteredOption(
        key = 'comp_pos.put_etc_before',
        defval = comp_pos_put_etc_before,
        doc =\
            "\n (bool)\n " +\
            "A bool (`False` by default) that determines whether " +\
            "or not components in `comp_pos.etc` should precede " +\
            "`comp_pos.interprov` and `comp_pos.intraprov` when " +\
            "applying corrections. If `True`, the precedence of " +\
            "components in `comp_pos.etc` will be between " +\
            "conventional positive components and interprovincial " +\
            "migrations. If `False`, it is " +\
            "set to follow interprovincial or intraprovincial " +\
            "migration, depending on the regional level of the " +\
            "population table.",
        validator = is_etc_p_b4_inters
    ),
    'method': RegisteredOption(
        key = "method",
        defval = method,
        doc =\
            "\n (str)\n    " +\
            "('1dist' by default) A method to apply " +\
            "counter-adjustments to a population table. Must be either " +\
            "one of '1dist', 'filler', or 'prop'.",
        validator = is_method
    ),
    'method_use.age_floor': RegisteredOption(
        key = 'method_use.age_floor',
        defval = method_use_age_floor,
        doc =\
            "\n (bool)\n    " +\
            "(True by default) If True, then the estime2 package " +\
            "uses `age.floor` option when calculating the minimum " +\
            "neighbouring age of the problematic record.",
        validator = is_bool_afloor
    ),
    'method_use.old_neigh': RegisteredOption(
        key = 'method_use.old_neigh',
        defval = method_use_old_neigh,
        doc =\
            "\n (bool)\n    " +\
            '(False by default) If True, then the estime2 package ' +\
            'considers ages older than the problematic age as ' +\
            'neighbouring ages. If False, then the package considers ' +\
            'ages younger than the problematic age as neighbouring ages.',
        validator = is_bool_old_neigh
    ),
    'method_use.second_pass': RegisteredOption(
        key = 'method_use.second_pass',
        defval = method_use_second_pass,
        doc =\
            "\n (bool)\n    " +\
            "(True by default) When using the '1dist' method, the " +\
            "correction method applies the same method once again " +\
            "to the same component in case where all the modification " +\
            "value cannot be fixed in one shot. " +\
            "Set it to True to use the second pass; set it to False " +\
            'to fix the negative value "as much as possible" and ' +\
            "nothing more.",
        validator = is_bool_mu2p
    ),
    'method_use.seq': RegisteredOption(
        key = 'method_use.seq',
        defval = method_use_seq,
        doc =\
            "(True by default) If True, then the one-distribution " +\
            "method will distribute the value of 1's to neighbouring " +\
            'ages and use the second pass (if applicable) ' +\
            '"sequentially". That is, when neighbouring ages are ' +\
            "given, then the process will look at the first " +\
            "`age.prop_size`-years of records and apply the " +\
            "one-distribution method and the second pass " +\
            "(if applicable) within those years. Then, it moves on " +\
            "to the next `age.prop_size`-years of records, does the " +\
            "same job, and continues until the minimum neighbouring " +\
            "age is met. If this option is False, then the " +\
            'one-distribution is applied "all at once" from the ' +\
            'maximum to the minimum neighbouring age, and the second ' +\
            'pass as well (if applicable).',
        validator = is_bool_seq
    ),
    'pop.age': RegisteredOption(
        key = "pop.age",
        defval = pop_age,
        doc =\
            "\n (str)\n    " +\
            'A name that denotes the age in a population table. ' +\
            'This option cannot be None.',
        validator = is_str_pop_age
    ),
    'pop.birth': RegisteredOption(
        key = "pop.birth",
        defval = pop_birth,
        doc =\
            "\n (str)\n    " +\
            'A name that denotes the birth in a population table. ' +\
            'This option cannot be None.',
        validator = is_str_pop_birth
    ),
    'pop.at_least': RegisteredOption(
        key = "pop.at_least",
        defval = pop_at_least,
        doc =\
            "\n (int)\n    " +\
            "A minimum end-of-period population estimate that needs\n" +\
            "to be kept." +\
            "This option cannot be None.",
        validator = is_pos_pop_at_least
    ),
    'pop.end': RegisteredOption(
        key = "pop.end",
        defval = pop_end,
        doc =\
            "\n (str or None)\n    " +\
            'A name that denotes the end-of-period population ' +\
            'in a population table.',
        validator = is_str_pop_end
    ),
    'pop.sex': RegisteredOption(
        key = "pop.sex",
        defval = pop_sex,
        doc =\
            "\n (str or None)\n    " +\
            'A name that denotes the sex in a population table. ' +\
            'If None, it assumes that records in a population table ' +\
            'all have the same sex.' +\
            'This option cannot be None.',
        validator = is_str_pop_sex
    ),
    'pop.start': RegisteredOption(
        key = "pop.start",
        defval = pop_start,
        doc =\
            "\n (str)\n    " +\
            'A name that denotes the start-of-period population ' +\
            'in a population table. ' +\
            'This option cannot be None.',
        validator = is_str_pop_start
    )
}



class CallableDynamicDoc:
    def __init__(self, func):
        self.__func__ = func

    def __call__(self, *args, **kwargs):
        return self.__func__(*args, **kwargs)

class DictWrapper:
    def __init__(self, d, prefix = ""):
        object.__setattr__(self, "d", d)
        object.__setattr__(self, "prefix", prefix)

    def __setattr__(self, key, val):
        prefix = object.__getattribute__(self, "prefix")
        if prefix:
            prefix += "."
        prefix += key
        if key in self.d and not isinstance(self.d[key], dict):
            _set_option(prefix, val)
        else:
            raise KeyError("Value cannot be set to nonexisting options.")

    def __getattr__(self, key):
        prefix = object.__getattribute__(self, "prefix")
        if prefix:
            prefix += "."
        prefix += key
        try:
            v = object.__getattribute__(self, "d")[key]
        except KeyError:
            raise KeyError("No such option exists.")
        if isinstance(v, dict):
            return DictWrapper(v, prefix)
        else:
            return _get_option(prefix)

    def __dir__(self):
        return list(self.d.keys())



def _get_option(pat: str):
    key = _get_single_key(pat)

    # walk the nested dict
    root, k = _get_root(key)
    return root[k]

def _get_registered_option(key: str):

    return _registered_options.get(key)

def _get_root(key: str):
    path = key.split(".")
    cursor = _global_config
    for p in path[:-1]:
        cursor = cursor[p]
    return cursor, path[-1]

def _get_single_key(pat: str):
    keys = _select_options(pat)
    if len(keys) == 0:
        raise KeyError("No such keys: {pat!r}".format(pat = pat))
    if len(keys) > 1:
        raise KeyError("Pattern matched multiple keys.")
    key = keys[0]
    
    return key

def _reset_option(*args):
    nargs = len(args)
    pat = None
    keys = None
    if nargs == 1:
        pat = args[0]
        keys = _select_options(pat)
    else:
        for patt in args:
            _reset_option(patt)
        return None

    if len(keys) == 0:
        raise KeyError("No such key(s)")

    if len(keys) > 1 and len(pat) < 4 and pat != "all":
        raise ValueError(
            "You must specify at least 4 characters when "
            "resetting multiple keys; use the special keyword "
            '"all" to reset all the options to their default value'
        )

    for k in keys:
        _set_option(k, _registered_options[k].defval)

def _select_options(pat: str):
    if pat in _registered_options:
        return [pat]
    keys = sorted(_registered_options.keys())
    if pat == 'all':
        return keys
    
    return [k for k in keys if re.search(pat, k, re.I)]

def _set_option(*args, **kwargs):
    nargs = len(args)
    if not nargs or nargs % 2 != 0:
        msg = "Must provide an even number of non-keyword arguments."
        raise ValueError(msg)
    silent = kwargs.pop("silent", False)
    if kwargs:
        msg2 = '_set_option() got an unexpected keyword argument "{kwarg}"'
        raise TypeError(msg2.format(list(kwargs.keys())[0]))
    for k, v in zip(args[::2], args[1::2]):
        key = _get_single_key(k)
        o = _get_registered_option(key)
        if o and o.validator:
            o.validator(v)
        
        root, k = _get_root(key)
        root[k] = v



class option_context:
    '''
    A context manager to temporarily set options in the `with` statment
    context. You need to invoke as:

    `option_context(pat, val, [(pat, val), ...])`

    Examples
    --------
    >>> with option_context('pop.at_least', 0, 'age.interval', [-5, 0]):
    ...
    '''


    def __init__(self, *args):
        if not (len(args) % 2 == 0 and len(args) >= 2):
            msg =\
                "Need to invoke as " +\
                "option_context(pat, val, [(pat, val), ...])."
            raise ValueError(msg)
        self.ops = list(zip(args[::2], args[1::2]))

    def __enter__(self):
        self.undo = [(pat, _get_option(pat)) for pat, val in self.ops]
        for pat, val in self.ops:
            _set_option(pat, val)

    def __exit__(self, *args):
        if self.undo:
            for pat, val in self.undo:
                _set_option(pat, val)



get_option = CallableDynamicDoc(_get_option)
options = DictWrapper(_global_config)
reset_option = CallableDynamicDoc(_reset_option)
set_option = CallableDynamicDoc(_set_option)




from estime2.age import Age
from estime2.config import (
    get_option,
    set_option
)
from estime2.poptable import (
    get_other_comp,
    get_L
)
from estime2.subpoptables import (
    apply_L,
    ProvPopTable,
    SubProvPopTable,
    return_goption_if_None
)
from estime2.poptable_resultswrapper import PopTableResultsWrapper
from functools import reduce
from typing import (
    List, 
    Optional,
    Union
)
import numpy as np
import pandas as pd



class AggregateProvPopTable():
    '''
    A list of `ProvPopTable`s.
    '''

    def __init__(
        self, 
        tbls: Union[dict, List[ProvPopTable]]
    ):
        '''
        Create an instance of AggregateProvPopTable.

        Usage
        -----
        AggregateProvPopTable(tbls: Union[dict, List[ProvPopTable]])

        Arguments
        ---------
        * `tbls`: a list of length at least 2, or a dict; a list/dict of
            `ProvPopTables` where each table has the same composition of 
            components.
        
        Details
        -------
        This class aims to fix negative populations in ProvPopTables of
        `tbls` by applying corrections to all the available components 
        in each table, including the interprovincial components. Dealing
        with interprovincial components in one table will involve 
        interprovincial components in other tables of `tbls`. 
        
        The following conditions apply when creating an instance of
        `AggregateProvPopTable`:

            1. The length of `tbls` should be at least 2.
            2. Each `ProvPopTable` in `tbls` should share the same name
                for the same component. 
            3. A component that exists in one table also has to exist in
                all the other tables as well. For example, if "Death" is
                one of the components in the table of `tbls`, then all the
                other tables in `tbls` must also have the "Death"
                component with the same name.
        '''

        assert len(tbls) >= 2, \
            "The `tbls` should contain at least 2 `ProvPopTable`s, " +\
            f"not {len(tbls)}."

        if isinstance(tbls, list):
            for i0, item in enumerate(tbls):
                if not isinstance(item, ProvPopTable):
                    raise TypeError(
                        f"The item of tbls at index {i0} "
                        f"is not a ProvPopTable but a {type(item)}."
                    )
            tbls = {k: tbls[k] for k in range(len(tbls))}
        elif isinstance(tbls, dict):
            for k, v in tbls.items():
                if not isinstance(v, ProvPopTable):
                    raise TypeError(
                        f"The value of tbls at key {k} "
                        f"is not a ProvPopTable but a {type(v)}."                
                    )
        else:
            raise NotImplementedError

        self.__tbls_dict = tbls
    
    def fix_all_issues(
        self, 
        method: Optional[str] = None,
        return_all_mods: bool = False
    ):
        '''
        Refer to the docstring of the `fix_all_issues(poptbls = self, 
        method = method, return_all_mods = return_all_mods)` function.
        '''

        return fix_all_issues(self, method, return_all_mods)

    def get_tbls_dict(self):
        '''
        Return `self.__tbls_dict`, the `dict` of all `ProvPopTable`s of
        `self`.

        Usage
        -----
        `self.get_tbls_dict()`
        '''

        return self.__tbls_dict

class AggregateSubProvPopTable():
    '''
    A list of `SubProvPopTable`s.
    '''

    def __init__(
        self, 
        tbls: Union[dict, List[SubProvPopTable]]
    ):
        '''
        Create an instance of AggregateSubProvPopTable.

        Usage
        -----
        AggregateSubProvPopTable(tbls: Union[dict, List[ProvPopTable]])

        Arguments
        ---------
        * `tbls`: a list of length at least 2; a list of 
            `SubProvPopTable`s where each table has the same composition of
            components.
        
        Details
        -------
        This class aims to fix negative populations in SubProvPopTables of
        `tbls` by applying corrections to all the available components 
        in each table, including the interprovincial components. Dealing
        with interprovincial components in one table will involve 
        interprovincial components in other tables of `tbls`. 
        
        The following conditions apply when creating an instance of
        `AggregateSubProvPopTable`:

            1. The length of `tbls` should be at least 2.
            2. Each `SubProvPopTable` in `tbls` should share the same 
                name for the same component. 
            3. A component that exists in one table also has to exist in
                all the other tables as well. For example, if "Death" is
                one of the components in the table of `tbls`, then all the
                other tables in `tbls` must also have the "Death"
                component with the same name.
        '''

        assert len(tbls) >= 2, \
            "The `lst_tbl` should contain at least 2 " +\
            "`SubProvPopTable`s, " +\
            f"not {len(tbls)}."

        if isinstance(tbls, list):
            for i0, item in enumerate(tbls):
                if not isinstance(item, SubProvPopTable):
                    raise TypeError(
                        f"The item of tbls at index {i0} "
                        f"is not a SubProvPopTable but a {type(item)}."
                    )
            tbls = {k: tbls[k] for k in range(len(tbls))}
        elif isinstance(tbls, dict):
            for k, v in tbls.items():
                if not isinstance(v, SubProvPopTable):
                    raise TypeError(
                        f"The value of tbls at key {k} "
                        f"is not a SubProvPopTable but a {type(v)}"
                    )
        else:
            raise NotImplementedError

        self.__tbls_dict = tbls

    def fix_all_issues(
        self, 
        method: Optional[str] = None,
        return_all_mods: bool = False
    ):
        '''
        Refer to the docstring of the `fix_all_issues(poptbls = self, 
        method = method, return_all_mods = return_all_mods)` function.
        '''

        return fix_all_issues(self, method, return_all_mods)

    def get_tbls_dict(self):
        '''
        Return `self.__tbls_dict`, the `dict` of all `SubProvPopTable`s of
        `self`.

        Usage
        -----
        `self.get_tbls_dict()`
        '''

        return self.__tbls_dict

def apply_Ns(
    tbl: Union[ProvPopTable, SubProvPopTable],
    t: int,
    comp: str,
    method: str,
    other_tbls: dict,
    comps_modifieds: dict
):
    '''
    Apply the modified L-vector of `tbl` (i.e. `comp_O`) to `tbl`, and the
    exact transferrables (i.e. `comp_N`s) of `other_tbls` to each of them. 
    Modify `comps_modifieds` whenever "comp_N" of each table actually 
    modifies itself.
    '''

    # Settings
    comp_L = f"{comp}_L"
    comp_N = f"{comp}_N"
    comp_O = f"{comp}_O"
    other_comp = get_other_comp(tbl, comp)
    other_comp_in_neg = other_comp in tbl.get_comp_neg()
    pop_groups = tbl.get_pop_groups()
    pop_age = pop_groups[1]
    comp_in_comp_end = comp in tbl.get_comp_end()
    to_transfer_ages = calculate_ages_to_transfer(
        tbl,
        comp,
        method,
        other_tbls
    )
    to_transfer_ages_pop_end = to_transfer_ages['age.to_transfer_pop_end']
    to_transfer_ages_comp = to_transfer_ages['age.to_transfer_comp']
    problematic_age = Age(to_transfer_ages_pop_end[-1])
    age_is_max = problematic_age.is_max()
    all_cols = tbl.columns.tolist()
    show_pop_end = tbl.get_pop_end() in all_cols
    is_subprov = tbl.is_subprov()

    # Get the L-vector of tbl
    L_df = get_L(tbl, comp, method, other_tbls)
    if (L_df[comp_L].values == 0).all():
        return tbl, other_tbls
    rearranging_col = [str(a) for a in L_df[pop_age].values]
    filtered_L_from_tbl = L_df[
        L_df[pop_age]\
            .apply(Age.get_showing_age)\
            .apply(str)\
            .isin([str(sc) for sc in to_transfer_ages_comp])
    ]

    # Get the sum of `comp_N`s of other_tbls
    exact_transferrables = get_Ns(tbl, comp, method, other_tbls)
    tbl_index = list(exact_transferrables.keys())
    sample_key = tbl_index[0]
    nrow = len(exact_transferrables[sample_key].index)
    comp_N_sum = np.repeat([0], nrow)
    for v in exact_transferrables.values():
        comp_N_sum += v[comp_N].values

    # Get the actual L vector (i.e. comp_O) to apply to tbl
    correction_to_tbl =\
        pd.DataFrame({
            'a': abs(filtered_L_from_tbl[comp_L].values),
            'b': abs(comp_N_sum)
        })\
        .min(axis = 1)\
        .values
    signs_for_receiver = calculate_signs(
        nrow, 
        age_is_max, 
        comp_in_comp_end, 
        other_comp_in_neg
    )
    signs_for_sender =\
        -signs_for_receiver if comp == other_comp else signs_for_receiver
    correction_to_tbl *= signs_for_sender
    if (correction_to_tbl == 0).all():
        return tbl, other_tbls
    filtered_L_from_tbl[comp_O] = correction_to_tbl
    comp_O_df = filtered_L_from_tbl.copy()
    del comp_O_df[comp_L]
    comp_O_df[pop_age] = comp_O_df[pop_age].apply(str)

    # Apply correction to tbl, and record any correction made to 
    # comps_modifieds[t].
    tbl_cp = tbl.copy()
    tbl_cp[pop_age] = tbl_cp[pop_age].apply(str)
    tbl_cp = tbl_cp.merge(comp_O_df, on = pop_groups, how = 'left')
    tbl_cp.fillna(0, inplace = True)
    tbl_cp[comp_O] = tbl_cp[comp_O].apply(int)
    if (tbl_cp[comp_O].values != 0).any():
        correction = tbl_cp[pop_groups + [comp_O]]
        df_list = []
        for rc in rearranging_col:
            df_list.append(correction[correction[pop_age] == rc])
        correction = pd.concat(df_list)
        comps_modifieds[t].append([correction, 'O', 'itself and other(s)'])
    tbl_cp[comp] += tbl_cp[comp_O]
    del tbl_cp[comp_O]
    if is_subprov:
        tbl_cp = SubProvPopTable(
            tbl_cp,
            reorder_cols = False,
            show_pop_end = show_pop_end,
            flag = False
        )
    else:
        tbl_cp = ProvPopTable(
            tbl_cp,
            reorder_cols = False,
            show_pop_end = show_pop_end,
            flag = False
        )

    # Apply correction to other_tbls, and record any correction made to
    # comps_modifieds[i], where i in tbl_index
    other_tbls_cp = other_tbls.copy()
    for i in tbl_index:
        other_tbl = other_tbls_cp[i].copy()
        other_tbl[pop_age] = other_tbl[pop_age].apply(str)
        comp_N_to_apply = exact_transferrables[i].copy()
        comp_N_to_apply[pop_age] = comp_N_to_apply[pop_age].apply(str)
        other_tbl = other_tbl\
            .merge(
                comp_N_to_apply, 
                on = pop_groups, 
                how = "left"
            )
        other_tbl.fillna(0, inplace = True)
        other_tbl[comp_N] = other_tbl[comp_N].apply(int)
        if (other_tbl[comp_N].values != 0).any():
            other_comp_P = f"{other_comp}_P"
            correction = other_tbl[pop_groups + [comp_N]]\
                .rename(columns = {comp_N: other_comp_P})
            df_list = []
            for rc in rearranging_col:
                df_list.append(correction[correction[pop_age] == rc])
            correction = pd.concat(df_list)
            comps_modifieds[i].append([correction, 'P', t])
        other_tbl[other_comp] += other_tbl[comp_N]
        del other_tbl[comp_N]
        if is_subprov:
            other_tbl = SubProvPopTable(
                other_tbl,
                reorder_cols = False,
                show_pop_end = show_pop_end,
                flag = False
            )
        else:
            other_tbl = ProvPopTable(
                other_tbl,
                reorder_cols = False,
                show_pop_end = show_pop_end,
                flag = False
            )
        other_tbls_cp[i] = other_tbl

    return tbl_cp, other_tbls_cp

def calculate_ages_to_transfer(
    tbl: Union[ProvPopTable, SubProvPopTable],
    comp: str,
    method: Optional[str],
    other_tbls: Optional[dict] = None
):
    '''
    Return the age(s) of end-of-period populations and `comp` to transfer 
    values from `tbl` (to other tbls) based on the `method` specified and
    absorbables based on `other_tbls`.

    Returns
    -------
    A `dict` of {`str`: `numpy.array` of `int`s}.
    '''

    comp_L = f"{comp}_L"
    comp_in_comp_end = comp in tbl.get_comp_end()
    pop_groups = tbl.get_pop_groups()
    pop_age = pop_groups[1]

    I = tbl.get_I()
    I = I.query('I != 0').sort_values(pop_groups)
    problematic = I.iloc[0, :]
    problematic_age = problematic[pop_age]
    age_max = get_option('age.max')
    age_is_max = problematic_age.is_max()

    # When computing L, it has to include other_tbls into consideration. 
    # That is:
    #     1. Apply get_J() to tables of other_tbls (using problematic_age
    #         from tbl) so that it computes maximum counter-adjustable 
    #         values for each table of other_tbls
    #     2. Sum up maximum counter-adjustables at each neighbouring age
    #     3. Use get_K() based on the sum of maximum counter-adjustables
    #         to compute exact counter-adjustables for tbl
    #     4. Use get_L() to get the combined DataFrame of exact 
    #         counter-adjustables and exact modifiables for tbl

    filtered_L =\
        get_L(tbl, comp, method, other_tbls).query(f"{comp_L} != 0")

    # to_transfer_ages_comp is `numpy.array` of `Age`s here
    to_transfer_ages_comp = filtered_L.copy()[pop_age].values
    if age_is_max and (not comp_in_comp_end):
        # If age_is_max, then old_neigh is disabled.
        to_transfer_ages_comp = np.append(
            to_transfer_ages_comp,
            np.array([Age(age_max), Age(age_max + 1)])
        )
        to_transfer_ages_comp = pd.unique(to_transfer_ages_comp)

    # to_transfer_ages_pop_end is also `numpy.array` of `Age`s here
    to_transfer_ages_pop_end = to_transfer_ages_comp.copy()
    if not comp_in_comp_end:
        to_transfer_ages_pop_end += 1
    to_transfer_ages_pop_end = pd.unique(to_transfer_ages_pop_end)

    # Conversion from `Age` to `int` required
    result = {}
    result['age.to_transfer_comp'] =\
        np.array([a1.get_showing_age() for a1 in to_transfer_ages_comp])
    result['age.to_transfer_pop_end'] =\
        np.array([a2.get_showing_age() for a2 in to_transfer_ages_pop_end])

    return result

def calculate_agg_rem(
    filtered_L_from_tbl: pd.DataFrame, 
    comp_Ms_sum: np.array, 
    age_is_max: bool, 
    comp_in_comp_end: bool,
    other_comp_in_neg: bool
):
    '''
    Calculate the "aggregate remainder". It is computed by first taking
    element-wise minimums of the L-column of `filtered_L_from_tbl` and 
    `comp_Ms_sum`, then some numbers change their signs depending on the 
    condition of `age_is_max`, `comp_in_comp_end`, and `other_comp_in_neg`.

    Returns
    -------
    A `numpy.array` of `int`s.
    '''

    comp_L_abs = abs(filtered_L_from_tbl.iloc[:, -1].values)
    comp_Ms_sum_abs = abs(comp_Ms_sum)
    agg_rem =\
        pd.DataFrame({
            'a': comp_L_abs,
            'b': comp_Ms_sum_abs
        })\
        .min(axis = 1)\
        .values
    signs = calculate_signs(
        len(agg_rem), 
        age_is_max, 
        comp_in_comp_end,
        other_comp_in_neg
    )
    agg_rem *= signs
    
    return agg_rem

def calculate_signs(
    length: int, 
    age_is_max: bool, 
    comp_in_comp_end: bool, 
    other_comp_in_neg: bool
):
    '''
    Return the `np.array` of 1's of length `length` with possibly different
    signs at each element based on information of `age_is_max`,
    `comp_in_comp_end`, and `other_comp_in_neg`.

    Preconditions
    -------------
    1. If `age_is_max` is `True` and `comp_in_comp_end` is `False`, then
        `length` must be greater than or equal to 3 (since the last two 
        slots of the returning `numpy.array` are reserved for modifiable 
        values, and counter-adjustments have to be made in other preceding
        slots).
    2. Otherwise, `length` must be greater than or equal to 2 (since the 
        last slot of the returning `numpy.array` are reserved for the
        modifiable value, and counter-adjustments have to be made in other
        preceding slots).

    Returns
    -------
    A `numpy.array` of `1`s and `-1`s.
    '''

    vec1 = np.repeat([1], length)
    vec1[-1] = -1
    if age_is_max and (not comp_in_comp_end):
        vec1[-2] = -1

    return -vec1 if other_comp_in_neg else vec1

def fix_all_issues(
    poptbls: Union[AggregateProvPopTable, AggregateSubProvPopTable],
    method: Optional[str] = None,
    return_all_mods: bool = False
):
    '''
    Apply modifications and counter-adjustments to `PopTable`s 
    of `self.__tbls_dict` so that there are no negative end-of-period 
    populations. In case of dealing with interprovincial components,
    "transfer and absorption" are also considered. Figures for 
    counter-adjustments in neighbouring ages are different based on 
    `method` specified. Set `return_all_mods` to `True` to get all the
    modifications and counter-adjustments made to each table of 
    `self.__tbls_dict`.

    Usage
    -----
    `fix_all_issues(
        poptbls: Union[AggregateProvPopTable, AggregateSubProvPopTable],
        method: Optional[str] = None,
        return_all_mods: bool = False
    )`

    Arguments
    ---------
    * `poptbls`: either an `AggregateProvPopTable` or 
        `AggregateSubProvPopTable` object
    * `method`: the same argument as in `PopTable.get_K()`
    * `return_all_mods`: a bool (`False` by default); if `True`, this 
        will return all the modifications and counter-adjustments made
        to each table of `self.__tbls_dict`.

    Returns
    -------
    If `return_all_mods` is `False`, then it returns a `dict` of fixed
    versions of `PopTable`s of `self.__tbls_dict`. Keys are 
    indices of those tables in `self.__tbls_dict`, and values are fixed
    tables.
    If `return_all_mods` is `True`, then regardless of whether or not
    there are corrections made to the tables, it returns a `dict` of 
    `PopTableResultsWrapper`s of each table of `self.__tbls_dict`. 
    Keys are indices of those tables in `self.__tbls_dict`, and values 
    are `PopTableResultsWrapper` of each table.
    '''

    tbls_dict = poptbls.get_tbls_dict().copy()
    tbls_dict_sample_key = list(tbls_dict.keys())[0]
    tbl_1st = tbls_dict[tbls_dict_sample_key]
    is_subprov = tbl_1st.is_subprov()
    poptbl_class = SubProvPopTable if is_subprov else ProvPopTable
    
    common_args = tbl_1st.get_args()
    prevs = {}
    for k, v in common_args.items():
        prevs[k] = get_option(k)
        set_option(k, v)
        
    comp_neg_to_use = tbl_1st.get_comp_neg()
    comp_pos_to_use = tbl_1st.get_comp_pos()
    comps = comp_neg_to_use + comp_pos_to_use

    pop_groups = tbl_1st.get_pop_groups()
    pop_sex = pop_groups[0]
    pop_age = pop_groups[1]
    all_cols = tbl_1st.columns.tolist()
    show_pop_end = tbl_1st.get_pop_end() in all_cols
    sexes = np.unique(tbl_1st[pop_sex])
    method = return_goption_if_None('method', method)

    comps_modifieds = {rl2: [] for rl2 in tbls_dict.keys()}
    muon_pal = [(False, 1), (True, 1), (False, 0), (True, 0)]

    # set_option(...) statement here shows that the predefined values of
    # `method_use.old_neigh` and `pop.at_least` will NOT be used.
    # That is, whatever you defined outside of this function will NOT be
    # used; only values in `muon_pal` will be used.
    set_option(
        'method_use.old_neigh', muon_pal[0][0],
        'pop.at_least', muon_pal[0][1]
    )
    fixed_tables = produce_fixed_tables(
        tbls_dict, 
        sexes, 
        pop_sex,
        pop_age,
        poptbl_class,
        is_subprov,
        show_pop_end,
        comps,
        method,
        comps_modifieds
    )
    still_has_neg = []
    for r in fixed_tables.keys():
        still_has_neg.append(
            (fixed_tables[r].calculate_pop().iloc[:, -1].values < 0).any()
        )
    has_neg = any(still_has_neg)
    mp = 1
    while has_neg and mp != len(muon_pal):
        set_option(
            'method_use.old_neigh', muon_pal[mp][0],
            'pop.at_least', muon_pal[mp][1]
        )
        fixed_tables = produce_fixed_tables(
            fixed_tables, 
            sexes, 
            pop_sex,
            pop_age,
            poptbl_class,
            is_subprov,
            show_pop_end,
            comps,
            method,
            comps_modifieds
        )
        still_has_neg = []
        for r in fixed_tables.keys():
            still_has_neg.append(
                (fixed_tables[r].calculate_pop().iloc[:, -1].values < 0)\
                .any()
            )
        has_neg = any(still_has_neg)
        mp += 1

    set_option(
        'method_use.old_neigh', muon_pal[0][0],
        'pop.at_least', muon_pal[0][1]
    )
    for k2, v2 in prevs.items():
        set_option(k2, v2)

    if return_all_mods:
        if method == '1dist':
            musp = get_option("method_use.second_pass")
            museq = get_option("method_use.seq")
            aps = get_option("age.prop_size")
            pass2 = ", 2nd pass" if musp else ""
            pass2seq =\
                f", {aps}-year sequence" if (musp and museq) else ""
            method = f"{method}{pass2}{pass2seq}"
        result = {
            rl4: PopTableResultsWrapper(
                orig_table = tbls_dict[rl4],
                result_fix_issues = [tbl4] + comps_modifieds[rl4],
                method = method
            ) \
            for rl4, tbl4 in fixed_tables.copy().items()
        }
        return result
    else:
        return fixed_tables

def get_Ms(
    tbl: Union[ProvPopTable, SubProvPopTable],
    comp: str,
    method: Optional[str],
    other_tbls: dict
):
    '''
    Compute the maximum amount that is transferrable from `comp` of `tbl`
    to the "other_comp" of `other_tbls` based on `method` specified.

    Arguments
    ---------
    * `tbl`: a `ProvPopTable` or `SubProvPopTable`
    * `comp`: a `str`; either:
        1. `tbl.get_interprov_out()` or `tbl.get_interprov_in()` if `tbl`
            is `ProvPopTable`
        2. any (negative or positive) component if `tbl` is 
            `SubProvPopTable`
    * `other_tbls`: a `dict` of:
        1. {`int`: `ProvPopTable`} if `tbl` is `ProvPopTable`
        2. {`int`: `SubProvPopTable`} if `tbl` is `SubProvPopTable`
    * `method`: a `str`; an argument of `tbl.get_L(comp, method)`

    Details
    -------
    1. If `tbl` is `ProvPopTable`: for a given interprovincial migration 
        component `comp`, the function computes the maximum transferrable 
        values for each `ProvPopTable` of `other_tbls`. That is, the 
        function computes how much of a possible correction to be made in 
        `tbl[comp]` (e.g. interprovincial migration OUT) is actually 
        transferrable (or "counter-counter-adjustable") using the opposite
        interprovincial migration components (e.g. interprovincial
        migration IN) of tables in `other_tbls`.
    2. If `tbl` is `SubProvPopTable`: for a given `tbl`, a 
        `SubProvPopTable`, the function computes the maximum transferrable
        values from `tbl[comp]` to:
            + `other_tbls[i][comp]` for an index i if `comp` is neither
                "intraprovincial migration OUT" nor "intraprovincial 
                migration IN", i.e. "other_comp" is the same as `comp` if 
                `tbl` is `SubProvPopTable` and `comp` is neither of the
                intraprovincial components of `tbl`;
            + `other_tbls[i][other_comp]` for an index i, where 
                `other_comp` is "intraprovincial migration IN" if `comp` is
                "intraprovincial migration OUT", and "intraprovincial 
                migration OUT" if `comp` is "intraprovincial migration IN",
                i.e. it behaves the same as `ProvPopTable` if `comp` is
                either of the intraprovincial components of `tbl`

    Returns
    -------
    Two `dict`s:
    1. A `dict` of {`int`: `pandas.DataFrame`}, where each DataFrame has 
        three columns: `pop.sex`, `pop.age`, and `comp_M` which stores 
        maximum transferrable values (at each record) from `tbl[comp]` to 
        `other_tbls[i][other_comp]` for an integer `i`;
    2. A `dict` of {`int`: `pandas.DataFrame`}, where each DataFrame has
        three columns: `pop.sex`, `pop.age`, and the column with the name
        "max_peal_0", which is computed as `max(pop.end - at.least, 0)`
        for each table.
    '''

    # Define constants
    to_transfer_ages = calculate_ages_to_transfer(
        tbl,
        comp,
        method,
        other_tbls
    )
    to_transfer_ages_pop_end = to_transfer_ages['age.to_transfer_pop_end']
    to_transfer_ages_comp = to_transfer_ages['age.to_transfer_comp']
    at_least = get_option('pop.at_least')
    pop_groups = tbl.get_pop_groups()
    pop_age = pop_groups[1]
    pop_end = tbl.get_pop_end()
    comp_L = f"{comp}_L"
    comp_M = f"{comp}_M"
    other_comp = get_other_comp(tbl, comp)
    problematic_age = Age(to_transfer_ages_pop_end[-1])
    comp_in_neg = comp in tbl.get_comp_neg()
    comp_in_comp_end = comp in tbl.get_comp_end()

    # Get exact amount to be corrected in tbl
    L_df = get_L(tbl, comp, method, other_tbls)
    filtered_L_from_tbl = L_df[
        L_df[pop_age]\
            .apply(Age.get_showing_age)\
            .apply(str)\
            .isin([str(sc) for sc in to_transfer_ages_comp])
    ]

    # Get respective end-of-period populations of other_tbls
    others_tbl_pop_end = {
        i0: tb0.calculate_pop()[
            tb0.calculate_pop()[pop_age]\
                .apply(Age.get_showing_age)\
                .apply(str)\
                .isin([str(sp) for sp in to_transfer_ages_pop_end])
        ] \
        for i0, tb0 in other_tbls.copy().items()        
    }
    otpe_sample_key = list(others_tbl_pop_end.keys())[0]
    no_need_for_rearranging =\
        (
            others_tbl_pop_end[otpe_sample_key][pop_age].values ==\
            to_transfer_ages_pop_end
        )\
        .all()
    if not no_need_for_rearranging:
        others_tbl_pop_end = rearrange_rows(
            others_tbl_pop_end, 
            to_transfer_ages_pop_end,
            pop_age
        )
    others_pop_end = {
        i1: tb1.assign(
            max_peal_0 = lambda df: df[pop_end].apply(
                lambda x: max(x - at_least, 0)
            )
        )\
        [pop_groups + ['max_peal_0']] \
        for i1, tb1 in others_tbl_pop_end.copy().items()
    }

    # Get respective correctable amounts in other_comp of other_tbls
    others_other_comp = {
        i2: tb2[
            tb2[pop_age]\
                .apply(Age.get_showing_age)\
                .apply(str)\
                .isin([str(sc) for sc in to_transfer_ages_comp])
        ]\
        [pop_groups + [other_comp]] \
        for i2, tb2 in other_tbls.copy().items()
    }
    if not no_need_for_rearranging:
        others_other_comp = rearrange_rows(
            others_other_comp,
            to_transfer_ages_comp,
            pop_age
        )
    max_transferrables_to_other_comp = {
        i2_1: tbl2_1[pop_groups] \
        for i2_1, tbl2_1 in others_other_comp.copy().items()
    }

    case_where_other_comp_is_neg = None
    if other_comp == comp:
        # subprov given
        # neither intraprov_out nor intraprov_in
        case_where_other_comp_is_neg = True if comp_in_neg else False
    else:
        # if prov, then comp is either interprov_out or interprov_in
        # if subprov, then comp is either intraprov_out or intraprov_in
        case_where_other_comp_is_neg = False if comp_in_neg else True    

    if problematic_age.is_max() and (not comp_in_comp_end):
        filtered_L_from_tbl_ca = filtered_L_from_tbl[comp_L].values[:-2]
        filtered_L_from_tbl_ms = filtered_L_from_tbl[comp_L].values[-2:]
        for i4 in others_pop_end.copy().keys():
            # other_tbl_max_peal_0_ca =\
            #     others_pop_end[i4]['max_peal_0'].values[:-1]
            other_tbl_max_peal_0_m =\
                others_pop_end[i4]['max_peal_0'].values[-1]
            other_tbl_other_comp_ca =\
                others_other_comp[i4][other_comp].values[:-2]
            other_tbl_other_comp_ms =\
                others_other_comp[i4][other_comp].values[-2:]
            transferrable_vals = get_Ms_case1(
                filtered_L_from_tbl_ca,
                filtered_L_from_tbl_ms,
                other_tbl_other_comp_ca,
                other_tbl_other_comp_ms,
                other_tbl_max_peal_0_m,
                case_where_other_comp_is_neg
            )
            max_transferrables_to_other_comp[i4][comp_M] =\
                transferrable_vals
    else:
        filtered_L_from_tbl_ca = filtered_L_from_tbl[comp_L].values[:-1]
        filtered_L_from_tbl_m = filtered_L_from_tbl[comp_L].values[-1]
        for i3 in others_pop_end.copy().keys():
            other_tbl_max_peal_0_m =\
                others_pop_end[i3]['max_peal_0'].values[-1]
            other_tbl_other_comp_ca =\
               others_other_comp[i3][other_comp].values[:-1]
            other_tbl_other_comp_m =\
                others_other_comp[i3][other_comp].values[-1]
            transferrable_vals = get_Ms_case2(
                filtered_L_from_tbl_ca,
                filtered_L_from_tbl_m,
                other_tbl_other_comp_ca,
                other_tbl_other_comp_m,
                other_tbl_max_peal_0_m,
                case_where_other_comp_is_neg
            )
            max_transferrables_to_other_comp[i3][comp_M] =\
                transferrable_vals

    return max_transferrables_to_other_comp, others_pop_end

def get_Ms_case1(
    filtered_L_from_tbl_ca,
    filtered_L_from_tbl_ms,
    other_tbl_other_comp_ca,
    other_tbl_other_comp_ms,
    other_tbl_max_peal_0_m,
    case_where_other_comp_is_neg
):
    '''
    Calculate the maximum transferable values when:
    
        1. the problematic age is the maximum age, AND;
        2. comp is NOT an end-of-period component.
    '''

    if case_where_other_comp_is_neg:
        # comp_in_neg if other_comp_is_comp
        # else comp in [interprov_in, intraprov_in]
        df_transferrable_ca = pd.DataFrame({
            'b': abs(filtered_L_from_tbl_ca),
            'c': other_tbl_other_comp_ca
        })
        transferrable_vals_at_ca =\
            -df_transferrable_ca.min(axis = 1).values
        abs_sum_trans_val_ca =\
            abs(sum(transferrable_vals_at_ca))
        a = min(
            other_tbl_max_peal_0_m,
            abs(filtered_L_from_tbl_ms[0]),
            abs_sum_trans_val_ca
        )
        b = min(
            other_tbl_max_peal_0_m - a,
            abs(filtered_L_from_tbl_ms[1]),
            abs_sum_trans_val_ca - a
        )
        top_items =\
            [0 for i6 in range(len(transferrable_vals_at_ca))]\
            if (a == 0 and b == 0) \
            else list(transferrable_vals_at_ca)
    else:
        # not comp_in_neg if other_comp_is_comp
        # else comp in [interprov_out, intraprov_out]
        b = -min(
            other_tbl_max_peal_0_m,
            abs(filtered_L_from_tbl_ms[1]),
            other_tbl_other_comp_ms[1]
        )
        a = -min(
            other_tbl_max_peal_0_m - abs(b),
            abs(filtered_L_from_tbl_ms[0]),
            other_tbl_other_comp_ms[0]
        )
        top_items =\
            [0 for i5 in range(len(filtered_L_from_tbl_ca))] \
            if (a == 0 and b == 0) \
            else list(abs(filtered_L_from_tbl_ca))

    return top_items + [a, b]

def get_Ms_case2(
    filtered_L_from_tbl_ca,
    filtered_L_from_tbl_m,
    other_tbl_other_comp_ca,
    other_tbl_other_comp_m,
    other_tbl_max_peal_0_m,
    case_where_other_comp_is_neg
):
    '''
    Calculate the maximum transferable values when:

        1. the problematic age is NOT the maximum age, OR;
        2. comp is an end-of-period component.
    '''

    if case_where_other_comp_is_neg:
        # comp_in_neg if other_comp_is_comp
        # else comp in [interprov_in, intraprov_in]
        df_transferrable_ca = pd.DataFrame({
            'b': abs(filtered_L_from_tbl_ca),
            'c': other_tbl_other_comp_ca
        })
        transferrable_vals_at_ca =\
            -df_transferrable_ca.min(axis = 1).values
        transferrable_val_at_m = min(
            other_tbl_max_peal_0_m,
            abs(filtered_L_from_tbl_m),
            abs(sum(transferrable_vals_at_ca))
        )
        top_items =\
            [0 for i8 in range(len(transferrable_vals_at_ca))]\
            if (transferrable_val_at_m == 0) \
            else list(transferrable_vals_at_ca)
    else:
        # not comp_in_neg if other_comp_is_comp
        # else comp in [interprov_out, intraprov_out]
        transferrable_val_at_m = -min(
            other_tbl_max_peal_0_m,
            abs(filtered_L_from_tbl_m),
            other_tbl_other_comp_m
        )
        top_items =\
            [0 for i7 in range(len(filtered_L_from_tbl_ca))]\
            if (transferrable_val_at_m == 0) \
            else list(abs(filtered_L_from_tbl_ca))

    return top_items + [transferrable_val_at_m]

def get_Ns(
    tbl: Union[ProvPopTable, SubProvPopTable],
    comp: str,
    method: Optional[str],
    other_tbls: dict
):
    '''
    Calculate exact transferrables from `tbl[comp]` to other tables of 
    `other_tbls`. Values are based on `tbl`'s "comp_L vector" computed by 
    `method`.

    Returns
    -------
    A `dict` of `{int: pandas.DataFrame}` where keys of returning `dict` is
    the same as `other_tbls.keys()`, and values are `pandas.DataFrame` with
    the following column names: `pop.sex`, `pop.age`, and `comp_N`. 
    `comp_N` stores the exact values of correction at each record for each
    table of `other_tbls`.
    '''

    comp_M = f"{comp}_M"
    comp_N = f"{comp}_N"
    other_comp = get_other_comp(tbl, comp)
    other_comp_in_neg = other_comp in tbl.get_comp_neg()
    pop_groups = tbl.get_pop_groups()
    pop_age = pop_groups[1]

    to_transfer_ages = calculate_ages_to_transfer(
        tbl,
        comp,
        method,
        other_tbls
    )
    to_transfer_ages_comp = to_transfer_ages['age.to_transfer_comp']
    L_df = get_L(tbl, comp, method, other_tbls)
    filtered_L_from_tbl = L_df[
        L_df[pop_age]\
            .apply(Age.get_showing_age)\
            .apply(str)\
            .isin([str(sc) for sc in to_transfer_ages_comp])
    ]

    max_transferrables, other_tbls_max_peal_0 = get_Ms(
        tbl, 
        comp, 
        method, 
        other_tbls
    )
    tbl_index, sample_key = sort_tbl_index(other_tbls_max_peal_0)
    problematic_age = other_tbls_max_peal_0[sample_key][pop_age].values[-1]
    age_is_max = problematic_age.is_max()
    comp_in_comp_end = comp in tbl.get_comp_end()
    nrow = len(max_transferrables[sample_key].index)
    comp_Ms_sum = np.repeat([0], nrow)
    for v2 in max_transferrables.copy().values():
        comp_Ms_sum += v2[comp_M].values

    # Calculate the aggregate remainder using filtered_L and comp_Ms_sum
    agg_rem = calculate_agg_rem(
        filtered_L_from_tbl, 
        comp_Ms_sum, 
        age_is_max, 
        comp_in_comp_end,
        other_comp_in_neg
    )

    # Generate comp_N for each table of other_tbls
    # Note: max_transferrables have values so that the sum of absolute
    # values of counter-adjustables is ALWAYS greater than or equal to
    # the sum of absolute values of modifiables, i.e. `flag` of the 
    # while-loop will NECESSARILY become False at some point.
    exact_transferrables = max_transferrables.copy()
    for i0 in tbl_index:
        exact_transferrables[i0][comp_N] = 0

    # signs_for_receiver = (+ -) if the receiver is positive:
    #     e.g.1. (not is_subprov) comp is IOM (neg) 
    #         so other_comp is IIM (pos);
    #     e.g.2. (is_subprov) comp is IMM (pos) 
    #         so other_comp is also IMM (pos)
    # signs_for_receiver = (- +) if the receiver is negative:
    #     e.g.1. (not is_subprov) comp is IIM (pos)
    #         so other_comp is IOM (neg);
    #     e.g.2. (is_subprov) comp is DTH (neg)
    #         so other_comp is also DTH (neg)
    signs_for_receiver = calculate_signs(
        nrow, 
        age_is_max, 
        comp_in_comp_end,
        other_comp_in_neg
    )

    flag = True
    while flag:
        if age_is_max and (not comp_in_comp_end):
            # method_use.old_neigh is ALWAYS False under this condition
            for i1 in tbl_index:
                comp_M_local = exact_transferrables[i1][comp_M].values
                df_local = pd.DataFrame({
                    'abs_agg_rem': abs(agg_rem),
                    'abs_comp_M_local': abs(comp_M_local)
                })
                agg_rem_local = df_local.min(axis = 1).values
                if (agg_rem_local[:-2] == 0).all():
                    agg_rem_local[-2:] = 0
                agg_rem_local *= signs_for_receiver
                comp_N_local = np.repeat([0], nrow)
                local_rem = abs(agg_rem_local[-2:])
                at_least_one_to_ca = False
                if local_rem[-1] >= 1:
                    comp_N_local[-1] = 1 if other_comp_in_neg else -1
                    at_least_one_to_ca = True
                elif local_rem[-2] >= 1:
                    comp_N_local[-2] = 1 if other_comp_in_neg else -1
                    at_least_one_to_ca = True
                
                if at_least_one_to_ca:
                    for j1 in range(nrow - 3, -1, -1):
                        if abs(agg_rem_local[j1]) >= 1:
                            comp_N_local[j1] =\
                                -1 if other_comp_in_neg else 1
                            break
                agg_rem -= comp_N_local
                exact_transferrables[i1][comp_M] -= comp_N_local
                exact_transferrables[i1][comp_N] += comp_N_local
            flag = (agg_rem[-2:] != [0, 0]).any()
        else:
            using_old_neigh =\
                to_transfer_ages_comp[0] > to_transfer_ages_comp[-1]
            for i2 in tbl_index:
                comp_M_local = exact_transferrables[i2][comp_M].values
                df_local = pd.DataFrame({
                    'abs_agg_rem': abs(agg_rem),
                    'abs_comp_M_local': abs(comp_M_local)
                })
                agg_rem_local = df_local.min(axis = 1).values
                if (agg_rem_local[:-1] == 0).all():
                    agg_rem_local[-1] = 0
                agg_rem_local *= signs_for_receiver
                comp_N_local = np.repeat([0], nrow)
                local_rem = abs(agg_rem_local[-1])
                if local_rem >= 1:
                    comp_N_local[-1] = 1 if other_comp_in_neg else -1
                    to_iterate = range(nrow - 1) if using_old_neigh \
                        else range(nrow - 2, -1, -1)
                    for j2 in to_iterate:
                        if abs(agg_rem_local[j2]) >= 1:
                            comp_N_local[j2] =\
                                -1 if other_comp_in_neg else 1
                            break
                agg_rem -= comp_N_local
                exact_transferrables[i2][comp_M] -= comp_N_local
                exact_transferrables[i2][comp_N] += comp_N_local
            flag = (agg_rem[-1] != 0)

    for i3 in tbl_index:
        del exact_transferrables[i3][comp_M]

    return exact_transferrables

def produce_fixed_tables(
    tbls_dict, 
    sexes, 
    pop_sex,
    pop_age,
    poptbl_class,
    is_subprov,
    show_pop_end,
    comps,
    method,
    comps_modifieds
):
    '''
    Produce fixed tables of `tbls_dict` and return those fixed_tables.
    Update `comps_modifieds` along the way.

    Arguments
    ---------
    * `tbls_dict`: a dict; the return value of 
        `Aggregate*PopTable.get_tbls_dict()`
    * `sexes`: a list-like; an array of all sexes in each table
    * `pop_sex`: a str; the name of sex column
    * `pop_age`: a str; the name of age column
    * `poptbl_class`: either `ProvPopTable` or `SubProvPopTable`
    * `is_subprov`: a bool; Is each table `SubProvPopTable`?
    * `show_pop_end`: a bool; Should each table show the end-of-period 
        populations?
    * `comps`: a list; a list of all components in each table
    * `method`: a str; one of "1dist", "filler", or "prop"
    * `comps_modifieds`: a dict of `{k: list-like}` where `k` is 
        `tbls_dict.keys()`
    '''

    fixed_tables = {
        rl1: {s: None for s in sexes} 
        for rl1 in tbls_dict.keys()
    }

    for sex in sexes:
        dict_all_tbls = {
            tt: [
                poptbl_class(
                    ttbl.copy().loc[lambda df: df[pop_sex] == sex],
                    reorder_cols = False, 
                    show_pop_end = show_pop_end,
                    flag = False
                )
            ] \
            for tt, ttbl in tbls_dict.items()
        }
        dat_keys = sort_tbl_index_simpler(dict_all_tbls)
        for t in dat_keys:
            # other_tbls is based on the result of previous 
            # iteration(s).
            other_tbls = dict_all_tbls.copy()
            tbl = other_tbls.pop(t)
            tbl = tbl[-1]
            interprovs = [
                tbl.get_interprov_out(),
                tbl.get_interprov_in()
            ]
            other_tbls = {
                t_other: tbl_other[-1] \
                for t_other, tbl_other in other_tbls.items()
            }
            flag = (sum(tbl.get_I()['I'].values != 0) != 0)
            i = 0
            num_zeros = sum(tbl.get_I()['I'].values == 0)
            masked_ages = []
            if flag:
                prob_ages =\
                    tbl.get_I().query('I != 0')[pop_age].values
                prob_ages = [str(a) for a in prob_ages.copy()]
            while flag:
                comp = comps[i]
                current_prob_age = str(
                    tbl.get_I().query('I != 0')[pop_age].values[0]
                )

                if is_subprov:
                    L_df = get_L(tbl, comp, method, other_tbls)
                    if (L_df.iloc[:, -1] != 0).any():
                        tbl, other_tbls = apply_Ns(
                            tbl,
                            t,
                            comp,
                            method,
                            other_tbls,
                            comps_modifieds
                        )
                else:
                    if comp in interprovs:
                        tbl, other_tbls = apply_Ns(
                            tbl,
                            t,
                            comp,
                            method,
                            other_tbls,
                            comps_modifieds
                        )
                    else:
                        tbl = apply_L(
                            tbl,
                            comp,
                            method,
                            comps_modifieds[t]
                        )
                
                I_vec = tbl.get_I()['I'].values
                new_num_zeros = sum(I_vec == 0)
                if all(I_vec == 0):
                    flag = False
                elif num_zeros != new_num_zeros:
                    i = 0
                    num_zeros = new_num_zeros
                elif i != (len(comps) - 1):
                    i += 1
                else:
                    # i.e. lack of values for corrections
                    # Correction could not be made to the problematic 
                    # record at the current problematic age.
                    if current_prob_age == prob_ages[-1]:
                        flag = False
                    else:
                        masked_ages.append(current_prob_age)
                        set_option('age.mask', masked_ages)
                        I_vec = tbl.get_I()['I'].values
                        new_num_zeros = sum(I_vec == 0)
                        i = 0
                        num_zeros = new_num_zeros
            set_option('age.mask', [])

            # Update dict_all_tbls after correcting tbl and 
            # (possibly) other_tbls
            for t1 in dat_keys:
                dict_all_tbls[t1].clear()
                if t1 == t:
                    dict_all_tbls[t1].append(tbl)
                else:
                    dict_all_tbls[t1].append(other_tbls[t1])

        # Update fixed_tables after applying corrections to all 
        # population tables of dict_all_tbls
        for t2 in fixed_tables.copy().keys():
            fixed_tables[t2][sex] = dict_all_tbls[t2][-1]

    # Bind rows of PopTables for each subdict in fixed_tables
    fixed_tables = {
        ind: reduce(
            lambda a, b: a.append(b, ignore_index = True),
            dict_tbl_by_sex.values()
        ) \
        for ind, dict_tbl_by_sex in fixed_tables.items()
    }

    # Convert DataFrames of fixed_tables into poptbl_class
    fixed_tables = {
        rl3: poptbl_class(
            tbl3.sort_values([pop_sex, pop_age]),
            reorder_cols = False, 
            show_pop_end = show_pop_end,
            flag = False
        ) \
        for rl3, tbl3 in fixed_tables.copy().items()
    }

    return fixed_tables

def rearrange_rows(dict_of_df, by, col):
    '''
    Rearrange rows of DataFrames in `dict_of_df` based on the array `by`
    and values in `df[col]`.

    Details
    -------
    Rows of each DataFrame in `dict_of_df` will be rearranged (or 
    reordered) by the values appeared in `by`. Rows will be reordered by
    matching values in `by` and the value in each `col`.
    '''

    result = {}
    for k, df in dict_of_df.copy().items():
        df_list = []
        for i in by:
            df_list.append(df[df[col] == i])
        result[k] = pd.concat(df_list)

    return result

def sort_tbl_index(other_tbls_pop_end: dict):
    '''
    Sort other_tbls_pop_end.keys() by:
        1. the end-of-period population at the problematic age, and;
        2. the sum of end-of-period populations of neighbouring ages
    in this order in descending orders.

    Returns
    -------
    Two items, in this order:
    1. A `numpy.array` of `int`s
    2. One of the keys of `other_tbls_pop_end`
    '''

    ind = []
    pe_problematic = []
    sum_nei = []
    sample_key = None
    for k, v in other_tbls_pop_end.items():
        ind.append(k)
        pe_problematic.append(v.iloc[:, -1].values[-1])
        sum_nei.append(v.iloc[:, -1].values[:-1].sum())
        sample_key = k

    index_tbl = {
        'ind': ind,
        'pe_problematic': pe_problematic,
        'sum_nei': sum_nei
    }
    index_tbl = pd.DataFrame(index_tbl)
    index_tbl = index_tbl\
        .sort_values(
            ['pe_problematic', 'sum_nei'],
            ascending = [False, False]
        )
    
    return index_tbl['ind'].values, sample_key

def sort_tbl_index_simpler(dict_all_tbls: dict):
    '''
    Sort keys of `dict_all_tbls` by the number of end-of-period population
    in the descending order.

    Returns
    -------
    A `list` of `object`s (dict_all_tbls.keys()).
    '''

    pops = {
        k: lst_v[-1].calculate_pop().iloc[:, -1].sum() \
        for k, lst_v in dict_all_tbls.items()
    }
    result_pops = {'keys': [], 'vals': []}
    for k2, v2 in pops.items():
        result_pops['keys'].append(k2)
        result_pops['vals'].append(v2)
    result_pops = pd.DataFrame(result_pops)
    result_pops = result_pops.sort_values(['vals'], ascending = [False])

    return result_pops['keys'].tolist()




from estime2.config import (
    get_option,
    set_option
)
from estime2.poptable import (
    PopTable,
    calculate_ages_to_modify_and_counter,
    return_goption_if_None
)
from estime2.poptable_resultswrapper import PopTableResultsWrapper
from typing import Collection as Axes
from typing import (
    List,
    Optional,
    Union
)
import numpy as np
import warnings

# pd.options.mode.chained_assignment = None
warnings.filterwarnings('ignore')
Dtype = Union[str, np.dtype, "ExtensionDtype"]



class ProvPopTable(PopTable):
    '''
    A table of populations and components at a provincial level.
    '''
    
    def __init__(
        self,
        data = None,
        index: Optional[Axes] = None,
        columns: Optional[Axes] = None,
        dtype: Optional[Dtype] = None,
        copy: bool = False,
        pop_sex: Optional[str] = None,
        pop_age: Optional[str] = None,
        pop_end: Optional[str] = None,
        pop_start: Optional[str] = None,
        pop_birth: Optional[str] = None,
        comp_neg_temp_out: Optional[str] = None,
        comp_neg_emi: Optional[str] = None,
        comp_neg_npr_out: Optional[str] = None,
        comp_neg_death: Optional[str] = None,
        comp_neg_interprov_out: Optional[str] = None,
        comp_neg_intraprov_out: Optional[str] = None,
        comp_neg_etc: Optional[List[str]] = None,
        comp_neg_put_etc_before: Optional[bool] = None,
        comp_pos_temp_in: Optional[str] = None,
        comp_pos_ret_emi: Optional[str] = None,
        comp_pos_npr_in: Optional[str] = None,
        comp_pos_immi: Optional[str] = None,
        comp_pos_interprov_in: Optional[str] = None,
        comp_pos_intraprov_in: Optional[str] = None,
        comp_pos_etc: Optional[List[str]] = None,
        comp_pos_put_etc_before: Optional[bool] = None,
        comp_end: Optional[List[str]] = None,
        reorder_cols: bool = True,
        show_pop_end: bool = False,
        is_subprov: bool = False,
        flag: bool = True
    ):

        # is_subprov must be False all the time
        # flag may be False if data is already grouped and summed by
        # pop.sex and pop.age
        if is_subprov:
            raise ValueError(
                "When creating an instance of ProvPopTable, "
                f"`is_subprov` must be False, and not {is_subprov}."
            )
        super().__init__(
            data, index, columns, dtype, copy,
            pop_sex,
            pop_age,
            pop_end,
            pop_start,
            pop_birth,
            comp_neg_temp_out,
            comp_neg_emi,
            comp_neg_npr_out,
            comp_neg_death,
            comp_neg_interprov_out,
            comp_neg_intraprov_out,
            comp_neg_etc,
            comp_neg_put_etc_before,
            comp_pos_temp_in,
            comp_pos_ret_emi,
            comp_pos_npr_in,
            comp_pos_immi,
            comp_pos_interprov_in,
            comp_pos_intraprov_in,
            comp_pos_etc,
            comp_pos_put_etc_before,
            comp_end,
            reorder_cols,
            show_pop_end,
            is_subprov,
            flag
        )

    def fix_issues(
        self, 
        method: Optional[str] = None, 
        return_all_mods: bool = False
    ):
        '''
        Refer to the docstring of the `fix_issues(poptbl = self, 
        method = method, return_all_mods = return_all_mods)` function.
        '''

        return fix_issues(self, method, return_all_mods)

class SubProvPopTable(PopTable):
    '''
    A table of populations and components at a subprovincial level.
    '''

    def __init__(
        self,
        data = None,
        index: Optional[Axes] = None,
        columns: Optional[Axes] = None,
        dtype: Optional[Dtype] = None,
        copy: bool = False,
        pop_sex: Optional[str] = None,
        pop_age: Optional[str] = None,
        pop_end: Optional[str] = None,
        pop_start: Optional[str] = None,
        pop_birth: Optional[str] = None,
        comp_neg_temp_out: Optional[str] = None,
        comp_neg_emi: Optional[str] = None,
        comp_neg_npr_out: Optional[str] = None,
        comp_neg_death: Optional[str] = None,
        comp_neg_interprov_out: Optional[str] = None,
        comp_neg_intraprov_out: Optional[str] = None,
        comp_neg_etc: Optional[List[str]] = None,
        comp_neg_put_etc_before: Optional[bool] = None,
        comp_pos_temp_in: Optional[str] = None,
        comp_pos_ret_emi: Optional[str] = None,
        comp_pos_npr_in: Optional[str] = None,
        comp_pos_immi: Optional[str] = None,
        comp_pos_interprov_in: Optional[str] = None,
        comp_pos_intraprov_in: Optional[str] = None,
        comp_pos_etc: Optional[List[str]] = None,
        comp_pos_put_etc_before: Optional[bool] = None,
        comp_end: Optional[List[str]] = None,
        reorder_cols: bool = True,
        show_pop_end: bool = False,
        is_subprov: bool = True,
        flag: bool = True
    ):
        '''
        Create an instance of SubProvPopTable. Note that every cell of the
        table except for those in the end-of-period population column
        should have non-negative values.

        Usage
        -----
        SubProvPopTable(
            data = None,
            index: Optional[Axes] = None,
            columns: Optional[Axes] = None,
            dtype: Optional[Dtype] = None,
            copy: bool = False,
            pop_sex: Optional[str] = None,
            pop_age: Optional[str] = None,
            pop_end: Optional[str] = None,
            pop_start: Optional[str] = None,
            pop_birth: Optional[str] = None,
            comp_neg_temp_out: Optional[str] = None,
            comp_neg_emi: Optional[str] = None,
            comp_neg_npr_out: Optional[str] = None,
            comp_neg_death: Optional[str] = None,
            comp_neg_interprov_out: Optional[str] = None,
            comp_neg_intraprov_out: Optional[str] = None,
            comp_neg_etc: Optional[List[str]] = None,
            comp_neg_put_etc_before: Optional[bool] = None,
            comp_pos_temp_in: Optional[str] = None,
            comp_pos_ret_emi: Optional[str] = None,
            comp_pos_npr_in: Optional[str] = None,
            comp_pos_immi: Optional[str] = None,
            comp_pos_interprov_in: Optional[str] = None,
            comp_pos_intraprov_in: Optional[str] = None,
            comp_pos_etc: Optional[List[str]] = None,
            comp_pos_put_etc_before: Optional[bool] = None,
            comp_end: Optional[List[str]] = None,
            reorder_cols: bool = True,
            show_pop_end: bool = False
        )

        Arguments
        ---------
        * `data`, `index`, `columns`, `dtype`, `copy`: inherited from 
            `pandas.DataFrame`
        * `pop_sex`, `pop_age`, `pop_end`, `pop_start`, `pop_birth`, 
            `comp_neg_temp_out`, `comp_neg_emi`, `comp_neg_npr_out`,
            `comp_neg_death`, `comp_neg_interprov_out`, 
            `comp_neg_intraprov_out`, `comp_neg_etc`,
            `comp_pos_temp_in`, `comp_pos_ret_emi`, `comp_pos_npr_in`,
            `comp_pos_immi`, `comp_pos_interprov_in`, 
            `comp_pos_intraprov_in`, `comp_pos_etc`,
            `comp_end`, `reorder_cols`, `show_pop_end`: inherited from
            `ProvPopTable`
        * `comp_neg_put_etc_before`: (`None` by default) a bool; if `True`,
            then components in `comp_neg.etc` will precede 
            `comp_neg.interprov_out` and `comp_neg.intraprov_out`, and 
            follow other conventional negative components. If `False`, 
            components will follow `comp_neg.intraprov_out` instead. 
            If `None`, then it uses a global option value 
            `estime2.options.comp_neg.put_etc_before` (`False` by default).
        * `comp_pos_put_etc_before`: (`None` by default) a bool; if `True`,
            then components in `comp_pos.etc` will precede 
            `comp_pos.interprov_in` and `comp_pos.intraprov_in`, and follow
            other conventional positive components. If `False`, components
            will follow `comp_pos.intraprov_in` instead. If `None`, then it
            uses a global option value 
            `estime2.options.comp_pos.put_etc_before` (`False` by default).
        
        Details
        -------
        1. `SubProvPopTable` is a special case of `ProvPopTable` where
            `is_subprov` is `True`.
        2. If `reorder_cols` is `True`, then the columns of `data` is
            reordered in the following order from the left. For a column
            corresponding to one of the `comp_*` arguments, it will be
            displayed in the table iff neither the argument nor the 
            corresponding global option is `None` or `[]`: 
                + `pop_sex`
                + `pop_age`
                + `pop_end` (if `show_pop_end` is True)
                + `pop_start`
                + `pop_birth`
                + `comp_neg_temp_out` 
                + `comp_neg_emi`
                + `comp_neg_npr_out`
                + `comp_neg_death`
                + `comp_neg_etc` (if given any & put_etc_before is True)
                + `comp_neg_interprov_out`
                + `comp_neg_intraprov_out`
                + `comp_neg_etc` (if given any & put_etc_before is False)
                + `comp_pos_temp_in`
                + `comp_pos_ret_emi`
                + `comp_pos_npr_in`
                + `comp_pos_immi`
                + `comp_pos_etc` (if given any & put_etc_before is True)
                + `comp_pos_interprov_in`
                + `comp_pos_intraprov_in`
                + `comp_pos_etc` (if given any & put_etc_before is False)
        3. Whether or not `reorder_cols` is `True`, the method looks for
            the component(s) responsible for the negative end-of-period
            population following the above order of `comp_*`'s.
        '''

        # is_subprov must be True all the time
        # flag may be False if data is already grouped and summed by
        # pop.sex and pop.age
        if not is_subprov:
            raise ValueError(
                "When creating an instance of SubProvPopTable, "
                f"`is_subprov` must be True, and not {is_subprov}."
            )
        super().__init__(
            data, index, columns, dtype, copy,
            pop_sex,
            pop_age,
            pop_end,
            pop_start,
            pop_birth,
            comp_neg_temp_out,
            comp_neg_emi,
            comp_neg_npr_out,
            comp_neg_death,
            comp_neg_interprov_out,
            comp_neg_intraprov_out,
            comp_neg_etc,
            comp_neg_put_etc_before,
            comp_pos_temp_in,
            comp_pos_ret_emi,
            comp_pos_npr_in,
            comp_pos_immi,
            comp_pos_interprov_in,
            comp_pos_intraprov_in,
            comp_pos_etc,
            comp_pos_put_etc_before,
            comp_end,
            reorder_cols,
            show_pop_end,
            is_subprov,
            flag
        )

    def fix_issues(
        self, 
        method: Optional[str] = None, 
        return_all_mods: bool = False
    ):
        '''
        Refer to the docstring of the `fix_issues(poptbl = self, 
        method = method, return_all_mods = return_all_mods)` function.
        '''

        return fix_issues(self, method, return_all_mods)

def apply_L(
    poptbl: Union[ProvPopTable, SubProvPopTable], 
    comp: str, 
    method: str, 
    comps_modified: List
):
    '''
    Add the return value of `poptbl.get_L(comp, method)` to the `comp`
    component of `poptbl`. Also, mutate `comps_modified` so that whenever
    the correction is applied (i.e. "L" is added) to `poptbl`, 
    `comps_modified` is appended with that correction "L".
    
    Usage
    -----
    `apply_L(
        poptbl: Union[ProvPopTable, SubProvPopTable], 
        comp: str, 
        method: str, 
        comps_modified: List
    )`

    Arguments
    ---------
    * `poptbl`: a `PopTable` with an unique sex. That is, only one sex
        exists in the table.
    * `comp`, `method`: `str`s; arguments for `poptbl.get_L(comp, method)`
    * `comps_modified`: a `list`; a list of `poptbl.get_L(comp, method)`
        whose "`comp`_L" column contains at least one nonzero value.
    '''

    pop_groups = poptbl.get_pop_groups()
    pop_age = pop_groups[1]
    all_cols = poptbl.columns.tolist()
    show_pop_end = poptbl.get_pop_end() in all_cols
    is_subprov = poptbl.is_subprov()

    comp_L = f"{comp}_L"
    L = poptbl.get_L(comp, method)
    if any(L[comp_L].values != 0):
        comps_modified.append([L, 'L', 'itself'])
    L[pop_age] = L[pop_age].apply(str)

    poptbl_cp = poptbl.copy()
    poptbl_cp[pop_age] = poptbl_cp[pop_age].apply(str)
    poptbl_cp = poptbl_cp.merge(L, on = pop_groups, how = "left")
    poptbl_cp.fillna(0, inplace = True)
    poptbl_cp[comp_L] = poptbl_cp[comp_L].apply(int)
    poptbl_cp[comp] += poptbl_cp[comp_L]
    del poptbl_cp[comp_L]

    if is_subprov:
        poptbl_cp = SubProvPopTable(
            poptbl_cp,
            reorder_cols = False, # since it is reordered already
            show_pop_end = show_pop_end,
            # flag = False since it is grouped and summed already
            flag = False     
        )
    else:
        poptbl_cp = ProvPopTable(
            poptbl_cp,
            reorder_cols = False, # since it is reordered already
            show_pop_end = show_pop_end,
            # flag = False since it is grouped and summed already
            flag = False          
        )

    return poptbl_cp

def fix_issues(
    poptbl: PopTable, 
    method: Optional[str] = None, 
    return_all_mods: bool = False
):
    '''
    Apply modifications and counter-adjustments to `poptbl` so that there
    are no negative end-of-period population. Figures for counter-
    adjustments in neighbouring ages are different based on `method`
    specified. Set `return_all_mods` to `True` to get all the 
    modifications and counter-adjustments made to `poptbl`.

    Usage
    -----
    `fix_issues(
        poptbl,
        method: Optional[str] = None,
        return_all_mods: bool = False
    )`

    Argument
    --------
    * `poptbl`: the `PopTable` object, either `ProvPopTable` or 
        `SubProvPopTable`
    * `method`: the same argument as in `poptbl.get_K(comp, method, J)`.
    * `return_all_mods`: a bool (`False` by default); if `True`, this 
        will return all the modifications and counter-adjustments made
        to `poptbl`.

    Details
    -------
    For each problematic record, it modifies problematic record(s) and
    makes counter-adjustments to records of neighbouring ages within a 
    component according to the `method` specified. For each component, 
    the modification and counter-adjustments made to that component of 
    `poptbl` will affect the end-of-period population. The process is 
    done until no new end-of-period population is negative.
    In case of `return_all_mods` is `True`, this method records how
    much of modification and counter-adjustments are made, and returns
    a `PopTableResultsWrapper` object.

    Returns
    -------
    If `return_all_mods` is `False`, then it returns a fixed version
    of `poptbl`, the current `poptbl`. 
    If `return_all_mods` is `True`, then regardless of whether or not
    there are corrections made to the original table, it returns a 
    `PopTableResultsWrapper`.
    '''

    poptbl_class = SubProvPopTable if poptbl.is_subprov() else ProvPopTable

    args_in_self = poptbl.get_args()
    prevs = {}
    for k, v in args_in_self.items():
        prevs[k] = get_option(k)
        set_option(k, v)
    
    comp_neg_to_use = poptbl.get_comp_neg()
    comp_pos_to_use = poptbl.get_comp_pos()
    comps = comp_neg_to_use + comp_pos_to_use

    pop_groups = poptbl.get_pop_groups()
    pop_sex = pop_groups[0]
    pop_age = pop_groups[1]
    all_cols = poptbl.columns.tolist()
    show_pop_end = poptbl.get_pop_end() in all_cols
    sexes = np.unique(poptbl[pop_sex])
    method = return_goption_if_None('method', method)

    comps_modified = []
    muon_pal = [(False, 1), (True, 1), (False, 0), (True, 0)]
    
    # set_option(...) statement here shows that the predefined values of
    # `method_use.old_neigh` and `pop.at_least` will NOT be used.
    # That is, whatever you defined outside of this function will NOT be
    # used; only values in `muon_pal` will be used.
    set_option(
        'method_use.old_neigh', muon_pal[0][0],
        'pop.at_least', muon_pal[0][1]
    )
    self_copy = produce_fixed_table(
        poptbl,
        sexes,
        pop_sex,
        pop_age,
        poptbl_class,
        show_pop_end,
        comps,
        method,
        comps_modified
    )
    has_neg = (self_copy.calculate_pop().iloc[:, -1].values < 0).any()
    mp = 1
    while has_neg and mp != len(muon_pal):
        set_option(
            'method_use.old_neigh', muon_pal[mp][0],
            'pop.at_least', muon_pal[mp][1]
        )
        self_copy = produce_fixed_table(
            self_copy,
            sexes,
            pop_sex,
            pop_age,
            poptbl_class,
            show_pop_end,
            comps,
            method,
            comps_modified
        )
        has_neg = (self_copy.calculate_pop().iloc[:, -1].values < 0).any()
        mp += 1

    set_option(
        'method_use.old_neigh', muon_pal[0][0],
        'pop.at_least', muon_pal[0][1]
    )
    for k2, v2 in prevs.items():
        set_option(k2, v2)

    if return_all_mods:
        if method == '1dist':
            musp = get_option("method_use.second_pass")
            museq = get_option("method_use.seq")
            aps = get_option("age.prop_size")
            pass2 = ", 2nd pass" if musp else ""
            pass2seq =\
                f", {aps}-year sequence" if (musp and museq) else ""
            method = f"{method}{pass2}{pass2seq}"
        result = PopTableResultsWrapper(
            orig_table = poptbl, 
            result_fix_issues = [self_copy] + comps_modified,
            method = method
        )
        return result
    else:
        return self_copy

def produce_fixed_table(
    poptbl,
    sexes,
    pop_sex,
    pop_age,
    poptbl_class,
    show_pop_end,
    comps,
    method,
    comps_modified
):
    '''
    Produce the fixed version of `poptbl` and return that fixed table.
    Update `comps_modified` along the way.

    Arguments
    ---------
    * `poptbl`: either the `ProvPopTable` or `SubProvPopTable` object
    * `sexes`: a list-like; an array of all sexes in each table
    * `pop_sex`: a str; the name of sex column
    * `pop_age`: a str; the name of age column
    * `poptbl_class`: either `ProvPopTable` or `SubProvPopTable`
    * `show_pop_end`: a bool; Should the returning table display the
            end-of-period populations?
    * `comps`: a list; a list of all components in each table
    * `method`: a str; one of "1dist", "filler", or "prop"
    * `comps_modified`: a list that is going to store all the corrections
        applied to `poptbl`
    '''

    self_copy = None
    for sex in sexes:
        self_cp_sex =\
            poptbl.copy().loc[lambda df: df[pop_sex] == sex]
        self_cp_sex = poptbl_class(
            self_cp_sex,
            reorder_cols = False, # since it is reordered already
            show_pop_end = show_pop_end,
            # flag = False since it is grouped and summed already
            flag = False
        )
        # `flag` is True iff `I` has at least one nonzero value
        flag = (sum(self_cp_sex.get_I()['I'].values != 0) != 0)
        i = 0
        num_zeros = sum(self_cp_sex.get_I()['I'].values == 0)
        masked_ages = []
        if flag:
            prob_ages =\
                self_cp_sex.get_I().query('I != 0')[pop_age].values
            prob_ages = [str(a) for a in prob_ages.copy()]
        while flag:
            comp = comps[i]
            current_prob_age = str(
                self_cp_sex.get_I().query('I != 0')[pop_age].values[0]
            )
            self_cp_sex = apply_L(
                self_cp_sex, 
                comp, 
                method,
                comps_modified
            )
            I_vec = self_cp_sex.get_I()['I'].values
            new_num_zeros = sum(I_vec == 0)
            if all(I_vec == 0):
                # all problematic records fixed
                flag = False # stop the while-loop iteration
            elif num_zeros != new_num_zeros:
                # one problematic record fixed
                i = 0 # move to the next problematic record
                num_zeros = new_num_zeros # assign a new number
            elif i != (len(comps) - 1): 
                # correction in progress
                i += 1
            else: 
                # i.e. lack of values for corrections
                # Correction could not be made to the problematic 
                # record at the current problematic age.
                if current_prob_age == prob_ages[-1]:
                    # The table `self_cp_sex` has no next problematic 
                    # record (i.e. the current problematic record is 
                    # the last problematic record in the table)
                    flag = False
                else:
                    # The table does have the next problematic record.
                    # Then, mask the current problematic record,
                    # recompute neighbouring ages and try again.
                    masked_ages.append(current_prob_age)
                    set_option('age.mask', masked_ages)
                    I_vec = self_cp_sex.get_I()['I'].values
                    new_num_zeros = sum(I_vec == 0)
                    i = 0
                    num_zeros = new_num_zeros
        set_option('age.mask', [])

        if self_copy is None:
            self_copy = self_cp_sex
        else:
            self_copy = self_copy\
                .append(self_cp_sex, ignore_index = True)

    self_copy = poptbl_class(
        self_copy.sort_values([pop_sex, pop_age]),
        reorder_cols = False, # since it is reordered already
        show_pop_end = show_pop_end,
        # flag = False since it is grouped and summed already
        flag = False
    )

    return self_copy




import estime2
import pytest



class TestCalculateAgesToModifyAndCounter:
    '''
    Test behaviours of the 
    `estime2.poptable.calculate_ages_to_modify_and_counter` function.
    '''

    @pytest.mark.parametrize(
        "problematic_age, comp_in_comp_end, expecteds",
        [
            (
                estime2.Age('100+'), False, {
                    'TT': {'age.to_modify': [99, 100], 'age.to_counter': [50, 98]},
                    'TF': {'age.to_modify': [99, 100], 'age.to_counter': [59, 98]},
                    'FT': {'age.to_modify': [99, 100], 'age.to_counter': [50, 98]},
                    'FF': {'age.to_modify': [99, 100], 'age.to_counter': [59, 98]}
                }
            ),
            (
                estime2.Age('100+'), True, {
                    'TT': {'age.to_modify': 100, 'age.to_counter': [50, 99]},
                    'TF': {'age.to_modify': 100, 'age.to_counter': [60, 99]},
                    'FT': {'age.to_modify': 100, 'age.to_counter': [50, 99]},
                    'FF': {'age.to_modify': 100, 'age.to_counter': [60, 99]}
                }
            ),
            (
                estime2.Age(99), False, {
                    'TT': {'age.to_modify': 98, 'age.to_counter': [99, 100]},
                    'TF': {'age.to_modify': 98, 'age.to_counter': [99, 100]},
                    'FT': {'age.to_modify': 98, 'age.to_counter': [50, 97]},
                    'FF': {'age.to_modify': 98, 'age.to_counter': [58, 97]}
                }
            ),
            (
                estime2.Age(99), True, {
                    'TT': {'age.to_modify': 99, 'age.to_counter': [100, 100]},
                    'TF': {'age.to_modify': 99, 'age.to_counter': [100, 100]},
                    'FT': {'age.to_modify': 99, 'age.to_counter': [50, 98]},
                    'FF': {'age.to_modify': 99, 'age.to_counter': [59, 98]}
                }
            ),
            (
                estime2.Age(97), False, {
                    'TT': {'age.to_modify': 96, 'age.to_counter': [97, 100]},
                    'TF': {'age.to_modify': 96, 'age.to_counter': [97, 100]},
                    'FT': {'age.to_modify': 96, 'age.to_counter': [50, 95]},
                    'FF': {'age.to_modify': 96, 'age.to_counter': [56, 95]}
                }
            ),
            (
                estime2.Age(97), True, {
                    'TT': {'age.to_modify': 97, 'age.to_counter': [98, 100]},
                    'TF': {'age.to_modify': 97, 'age.to_counter': [98, 100]},
                    'FT': {'age.to_modify': 97, 'age.to_counter': [50, 96]},
                    'FF': {'age.to_modify': 97, 'age.to_counter': [57, 96]}
                }
            ),
            (
                estime2.Age(55), False, {
                    'TT': {'age.to_modify': 54, 'age.to_counter': [55, 94]},
                    'TF': {'age.to_modify': 54, 'age.to_counter': [55, 94]},
                    'FT': {'age.to_modify': 54, 'age.to_counter': [14, 53]},
                    'FF': {'age.to_modify': 54, 'age.to_counter': [14, 53]}
                }
            ),
            (
                estime2.Age(55), True, {
                    'TT': {'age.to_modify': 55, 'age.to_counter': [56, 95]},
                    'TF': {'age.to_modify': 55, 'age.to_counter': [56, 95]},
                    'FT': {'age.to_modify': 55, 'age.to_counter': [15, 54]},
                    'FF': {'age.to_modify': 55, 'age.to_counter': [15, 54]}
                }
            ),
            (
                estime2.Age(45), False, {
                    'TT': {'age.to_modify': 44, 'age.to_counter': [45, 84]},
                    'TF': {'age.to_modify': 44, 'age.to_counter': [45, 84]},
                    'FT': {'age.to_modify': 44, 'age.to_counter': [4, 43]},
                    'FF': {'age.to_modify': 44, 'age.to_counter': [4, 43]}
                }
            ),
            (
                estime2.Age(45), True, {
                    'TT': {'age.to_modify': 45, 'age.to_counter': [46, 85]},
                    'TF': {'age.to_modify': 45, 'age.to_counter': [46, 85]},
                    'FT': {'age.to_modify': 45, 'age.to_counter': [5, 44]},
                    'FF': {'age.to_modify': 45, 'age.to_counter': [5, 44]}
                }
            ),
            (
                estime2.Age(7), False, {
                    'TT': {'age.to_modify': 6, 'age.to_counter': [7, 50]},
                    'TF': {'age.to_modify': 6, 'age.to_counter': [7, 46]},
                    'FT': {'age.to_modify': 6, 'age.to_counter': [-1, 5]},
                    'FF': {'age.to_modify': 6, 'age.to_counter': [-1, 5]}
                }
            ),
            (
                estime2.Age(7), True, {
                    'TT': {'age.to_modify': 7, 'age.to_counter': [8, 50]},
                    'TF': {'age.to_modify': 7, 'age.to_counter': [8, 47]},
                    'FT': {'age.to_modify': 7, 'age.to_counter': [0, 6]},
                    'FF': {'age.to_modify': 7, 'age.to_counter': [0, 6]}
                }
            ),
            (
                estime2.Age(1), False, {
                'TT': {'age.to_modify': 0, 'age.to_counter': [1, 50]},
                'TF': {'age.to_modify': 0, 'age.to_counter': [1, 40]},
                'FT': {'age.to_modify': 0, 'age.to_counter': [-1, -1]},
                'FF': {'age.to_modify': 0, 'age.to_counter': [-1, -1]}
                }
            ),
            (
                estime2.Age(1), True, {
                'TT': {'age.to_modify': 1, 'age.to_counter': [2, 50]},
                'TF': {'age.to_modify': 1, 'age.to_counter': [2, 41]},
                'FT': {'age.to_modify': 1, 'age.to_counter': [0, 0]},
                'FF': {'age.to_modify': 1, 'age.to_counter': [0, 0]}
                }
            ),
            (
                estime2.Age(0), False, {
                'TT': {'age.to_modify': -1, 'age.to_counter': [0, 50]},
                'TF': {'age.to_modify': -1, 'age.to_counter': [0, 39]},
                'FT': {'age.to_modify': -1, 'age.to_counter': [0, 50]},
                'FF': {'age.to_modify': -1, 'age.to_counter': [0, 39]}
                }
            ),
            (
                estime2.Age(0), True, {
                'TT': {'age.to_modify': 0, 'age.to_counter': [1, 50]},
                'TF': {'age.to_modify': 0, 'age.to_counter': [1, 40]},
                'FT': {'age.to_modify': 0, 'age.to_counter': [1, 50]},
                'FF': {'age.to_modify': 0, 'age.to_counter': [1, 40]}
                }
            )
        ]
    )
    def test_ages_int(
        self, problematic_age, comp_in_comp_end, expecteds
    ):
        '''
        Test if the function correctly calculates modification age and
        neighbouring ages of ages depending on comp_in_comp_end, and the 
        following global options:

            + `method_use.old_neigh` == True or False
            + `method_use.age_floor` == True or False
            + `age.floor` == 50
            + `age.down_to` == 40
            + `age.max` == 99

        In case where the problematic age is either 0 or 100+, test if the
        function overrides the predefined `method_use.old_neigh` option and
        applies an appropriate approach:

            + If the age is 0, then the function behaves as if 
                `method_use.old_neigh` is True REGARDLESS of the original
                global option value;
            + If the age is 100+, then the function behaves as if
                `method_use.old_neigh` is False REGARDLESS of the original
                global option value
        '''

        keys = ['TT', 'TF', 'FT', 'FF']
        bool_selected = [False, True]
        for ii in keys:
            estime2.set_option(
                'method_use.old_neigh', bool_selected[ii[0] == 'T'],
                'method_use.age_floor', bool_selected[ii[1] == 'T']
            )
            result = estime2.poptable.calculate_ages_to_modify_and_counter(
                problematic_age, 
                comp_in_comp_end
            )
            assert expecteds[ii] == result
